@@ -16,6 +16,9 @@ from localstack.services.dynamodbstreams import dynamodbstreams_api
 # cache table definitions - used for testing
 TABLE_DEFINITIONS = {}
 
+# cache table taggings
+TABLE_TAGS = {}
+
 # action header prefix
 ACTION_PREFIX = 'DynamoDB_20120810'
 
@@ -41,6 +44,7 @@ class ProxyListenerDynamoDB(ProxyListener):
                 headers = {'Refresh': '0; url=%s/shell/' % config.TEST_DYNAMODB_URL}
                 return aws_responses.requests_response('', headers=headers)
             return True
+
         if method == 'OPTIONS':
             return 200
 
@@ -60,6 +64,7 @@ class ProxyListenerDynamoDB(ProxyListener):
             table_names = ddb_client.list_tables()['TableNames']
             if to_str(data['TableName']) in table_names:
                 return 200
+
         elif action in ('%s.PutItem' % ACTION_PREFIX, '%s.UpdateItem' % ACTION_PREFIX, '%s.DeleteItem' % ACTION_PREFIX):
             # find an existing item and store it in a thread-local, so we can access it in return_response,
             # in order to determine whether an item already existed (MODIFY) or not (INSERT)
@@ -69,21 +74,25 @@ class ProxyListenerDynamoDB(ProxyListener):
                 if 'ResourceNotFoundException' in str(e):
                     return get_table_not_found_error()
                 raise
+
             # Fix incorrect values if ReturnValues==ALL_OLD and ReturnConsumedCapacity is
             # empty, see https://github.com/localstack/localstack/issues/2049
             if data.get('ReturnValues') == 'ALL_OLD' and not data.get('ReturnConsumedCapacity'):
                 data['ReturnConsumedCapacity'] = 'TOTAL'
                 return Request(data=json.dumps(data), method=method, headers=headers)
+
         elif action == '%s.DescribeTable' % ACTION_PREFIX:
             # Check if table exists, to avoid error log output from DynamoDBLocal
             table_names = ddb_client.list_tables()['TableNames']
             if to_str(data['TableName']) not in table_names:
                 return get_table_not_found_error()
+
         elif action == '%s.DeleteTable' % ACTION_PREFIX:
             # Check if table exists, to avoid error log output from DynamoDBLocal
             table_names = ddb_client.list_tables()['TableNames']
             if to_str(data['TableName']) not in table_names:
                 return get_table_not_found_error()
+
         elif action == '%s.BatchWriteItem' % ACTION_PREFIX:
             existing_items = []
             for table_name in sorted(data['RequestItems'].keys()):
@@ -93,6 +102,7 @@ class ProxyListenerDynamoDB(ProxyListener):
                         if inner_request:
                             existing_items.append(find_existing_item(inner_request, table_name))
             ProxyListenerDynamoDB.thread_local.existing_items = existing_items
+
         elif action == '%s.TransactWriteItems' % ACTION_PREFIX:
             existing_items = []
             for item in data['TransactItems']:
@@ -101,6 +111,7 @@ class ProxyListenerDynamoDB(ProxyListener):
                     if inner_item:
                         existing_items.append(find_existing_item(inner_item))
             ProxyListenerDynamoDB.thread_local.existing_items = existing_items
+
         elif action == '%s.UpdateTimeToLive' % ACTION_PREFIX:
             # TODO: TTL status is maintained/mocked but no real expiry is happening for items
             response = Response()
@@ -112,6 +123,7 @@ class ProxyListenerDynamoDB(ProxyListener):
             response._content = json.dumps({'TimeToLiveSpecification': data['TimeToLiveSpecification']})
             fix_headers_for_updated_response(response)
             return response
+
         elif action == '%s.DescribeTimeToLive' % ACTION_PREFIX:
             response = Response()
             response.status_code = 200
@@ -128,18 +140,26 @@ class ProxyListenerDynamoDB(ProxyListener):
                 })
             else:  # TTL for dynamodb table not set
                 response._content = json.dumps({'TimeToLiveDescription': {'TimeToLiveStatus': 'DISABLED'}})
+
             fix_headers_for_updated_response(response)
             return response
+
         elif action == '%s.TagResource' % ACTION_PREFIX or action == '%s.UntagResource' % ACTION_PREFIX:
             response = Response()
             response.status_code = 200
             response._content = ''  # returns an empty body on success.
             fix_headers_for_updated_response(response)
             return response
+
         elif action == '%s.ListTagsOfResource' % ACTION_PREFIX:
             response = Response()
             response.status_code = 200
-            response._content = json.dumps({'Tags': []})  # TODO: mocked and returns an empty list of tags for now.
+            response._content = json.dumps({
+                'Tags': [
+                    {'Key': k, 'Value': v}
+                    for k, v in TABLE_TAGS.get(data['ResourceArn'], {}).items()
+                ]
+            })
             fix_headers_for_updated_response(response)
             return response
 
@@ -148,6 +168,7 @@ class ProxyListenerDynamoDB(ProxyListener):
     def return_response(self, method, path, data, headers, response):
         if path.startswith('/shell') or method == 'GET':
             return
+
         data = json.loads(to_str(data))
 
         # update table definitions
@@ -156,9 +177,11 @@ class ProxyListenerDynamoDB(ProxyListener):
 
         if response._content:
             # fix the table and latest stream ARNs (DynamoDBLocal hardcodes "ddblocal" as the region)
-            content_replaced = re.sub(r'("TableArn"|"LatestStreamArn"|"StreamArn")\s*:\s*"arn:aws:dynamodb:' +
-            'ddblocal:([^"]+)"', r'\1: "arn:aws:dynamodb:%s:\2"' % aws_stack.get_region(),
-            to_str(response._content))
+            content_replaced = re.sub(
+                r'("TableArn"|"LatestStreamArn"|"StreamArn")\s*:\s*"arn:aws:dynamodb:ddblocal:([^"]+)"',
+                r'\1: "arn:aws:dynamodb:%s:\2"' % aws_stack.get_region(),
+                to_str(response._content)
+            )
             if content_replaced != response._content:
                 response._content = content_replaced
                 fix_headers_for_updated_response(response)
@@ -191,8 +214,10 @@ class ProxyListenerDynamoDB(ProxyListener):
                 record['dynamodb']['SizeBytes'] = len(json.dumps(updated_item))
         elif action == '%s.BatchWriteItem' % ACTION_PREFIX:
             records = self.prepare_batch_write_item_records(record, data)
+
         elif action == '%s.TransactWriteItems' % ACTION_PREFIX:
             records = self.prepare_transact_write_item_records(record, data)
+
         elif action == '%s.PutItem' % ACTION_PREFIX:
             if response.status_code == 200:
                 existing_item = self._thread_local('existing_item')
@@ -210,6 +235,7 @@ class ProxyListenerDynamoDB(ProxyListener):
                 record['dynamodb']['SizeBytes'] = len(json.dumps(data['Item']))
                 if existing_item:
                     record['dynamodb']['OldImage'] = existing_item
+
         elif action in ['%s.GetItem' % ACTION_PREFIX, '%s.Query' % ACTION_PREFIX]:
             if response.status_code == 200:
                 content = json.loads(to_str(response.content))
@@ -224,30 +250,62 @@ class ProxyListenerDynamoDB(ProxyListener):
                     }
                     response._content = json.dumps(content)
                     fix_headers_for_updated_response(response)
+
         elif action == '%s.DeleteItem' % ACTION_PREFIX:
             if response.status_code == 200:
                 old_item = self._thread_local('existing_item')
                 record['eventName'] = 'REMOVE'
                 record['dynamodb']['Keys'] = data['Key']
                 record['dynamodb']['OldImage'] = old_item
+
         elif action == '%s.CreateTable' % ACTION_PREFIX:
             if 'StreamSpecification' in data:
                 if response.status_code == 200:
                     content = json.loads(to_str(response._content))
                     create_dynamodb_stream(data, content['TableDescription'].get('LatestStreamLabel'))
+
             event_publisher.fire_event(event_publisher.EVENT_DYNAMODB_CREATE_TABLE,
                 payload={'n': event_publisher.get_hash(data['TableName'])})
+
+            if data.get('Tags'):
+                table_arn = json.loads(response._content)['TableDescription']['TableArn']
+                TABLE_TAGS[table_arn] = {tag['Key']: tag['Value'] for tag in data['Tags']}
+
             return
+
         elif action == '%s.DeleteTable' % ACTION_PREFIX:
-            event_publisher.fire_event(event_publisher.EVENT_DYNAMODB_DELETE_TABLE,
-                payload={'n': event_publisher.get_hash(data['TableName'])})
+            event_publisher.fire_event(
+                event_publisher.EVENT_DYNAMODB_DELETE_TABLE,
+                payload={'n': event_publisher.get_hash(data['TableName'])}
+            )
+
+            TABLE_TAGS.pop(json.loads(response._content).get('TableDescription', {}).get('TableArn', None), None)
+
             return
+
         elif action == '%s.UpdateTable' % ACTION_PREFIX:
             if 'StreamSpecification' in data:
                 if response.status_code == 200:
                     content = json.loads(to_str(response._content))
                     create_dynamodb_stream(data, content['TableDescription'].get('LatestStreamLabel'))
             return
+
+        elif action == '%s.TagResource' % ACTION_PREFIX:
+            table_arn = data['ResourceArn']
+            if table_arn not in TABLE_TAGS:
+                TABLE_TAGS[table_arn] = {}
+
+            TABLE_TAGS[table_arn].update({tag['Key']: tag['Value'] for tag in data.get('Tags', [])})
+
+            return
+
+        elif action == '%s.UntagResource' % ACTION_PREFIX:
+            table_arn = data['ResourceArn']
+            for tag_key in data.get('TagKeys', []):
+                TABLE_TAGS.get(table_arn, {}).pop(tag_key, None)
+
+            return
+
         else:
             # nothing to do
             return
@@ -255,6 +313,7 @@ class ProxyListenerDynamoDB(ProxyListener):
         if len(records) > 0 and 'eventName' in records[0]:
             if 'TableName' in data:
                 records[0]['eventSourceARN'] = aws_stack.dynamodb_table_arn(data['TableName'])
+
             forward_to_lambda(records)
             forward_to_ddb_stream(records)
 
@@ -339,17 +398,14 @@ class ProxyListenerDynamoDB(ProxyListener):
                 records.append(new_record)
         return records
 
-    def _thread_local(self, name, default=None):
+    @staticmethod
+    def _thread_local(name, default=None):
         try:
             return getattr(ProxyListenerDynamoDB.thread_local, name)
         except AttributeError:
             return default
 
 
-# instantiate listener
-UPDATE_DYNAMODB = ProxyListenerDynamoDB()
-
-
 def find_existing_item(put_item, table_name=None):
     table_name = table_name or put_item['TableName']
     ddb_client = aws_stack.connect_to_service('dynamodb')
@@ -401,11 +457,17 @@ def calculate_crc32(response):
 def create_dynamodb_stream(data, latest_stream_label):
     stream = data['StreamSpecification']
     enabled = stream.get('StreamEnabled')
+
     if enabled not in [False, 'False']:
         table_name = data['TableName']
         view_type = stream['StreamViewType']
-        dynamodbstreams_api.add_dynamodb_stream(table_name=table_name,
-            latest_stream_label=latest_stream_label, view_type=view_type, enabled=enabled)
+
+        dynamodbstreams_api.add_dynamodb_stream(
+            table_name=table_name,
+            latest_stream_label=latest_stream_label,
+            view_type=view_type,
+            enabled=enabled
+        )
 
 
 def forward_to_lambda(records):
@@ -427,12 +489,17 @@ def dynamodb_extract_keys(item, table_name):
     if table_name not in TABLE_DEFINITIONS:
         LOGGER.warning('Unknown table: %s not found in %s' % (table_name, TABLE_DEFINITIONS))
         return None
+
     for key in TABLE_DEFINITIONS[table_name]['KeySchema']:
         attr_name = key['AttributeName']
         if attr_name not in item:
-            return error_response(error_type='ValidationException',
-                message='One of the required keys was not given a value')
+            return error_response(
+                error_type='ValidationException',
+                message='One of the required keys was not given a value'
+            )
+
         result[attr_name] = item[attr_name]
+
     return result
 
 
@@ -458,3 +525,7 @@ def error_response_throughput():
             'Consider increasing your provisioning level with the UpdateTable API')
     error_type = 'ProvisionedThroughputExceededException'
     return error_response(message, error_type)
+
+
+# instantiate listener
+UPDATE_DYNAMODB = ProxyListenerDynamoDB()
